#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include "types.h"
#include "gnu_mask.xbm"
#include "gnu_col.xpm"

int init(int xsz, int ysz);
void clean_up(void);
void set_no_decoration(Window win);

Display *dpy;
int scr;
Window win, rwin;
XImage *ximg;
Pixmap pixmap, mask;

int main(void) {
	int wmapped = 0;
	int done = 0;
	
	if(init(gnu_mask_width, gnu_mask_height) == -1) {
		return EXIT_FAILURE;
	}

	while(!done) {
		XEvent event;
		XNextEvent(dpy, &event);

		switch(event.type) {
		case MapNotify:
			wmapped = 1;
			break;

		case UnmapNotify:
			wmapped = 0;
			break;

		case Expose:
			if(wmapped && event.xexpose.count == 0) {
				GC gc = XCreateGC(dpy, win, 0, 0);
				XCopyArea(dpy, pixmap, win, gc, 0,  0, gnu_mask_width, gnu_mask_height, 0, 0);
				XFreeGC(dpy, gc);
				XFlush(dpy);
			}
			break;

		case KeyPress:
			{
				KeySym keysym = XLookupKeysym((XKeyEvent*)&event.xkey, 0);
				if((keysym & 0xff) == 27) {
					done = 1;
				}
			}
			break;

		case ClientMessage:
			{
				char *atom_name = XGetAtomName(dpy, event.xclient.message_type);
				if(!strcmp(atom_name, "WM_PROTOCOLS")) {
					done = 1;
				}
				XFree(atom_name);
			}
			break;

		default:
			break;
		}
	}

	clean_up();

	return 0;
}

int init(int xsz, int ysz) {
	XSetWindowAttributes xattr;
	Atom wm_del;
	XTextProperty wname;
	XClassHint chint;
	int event_base, error_base;
	char *title = "Shaped window";

	if(!(dpy = XOpenDisplay(0))) {
		fputs("Failed to connect to the X server\n", stderr);
		return -1;
	}

	scr = DefaultScreen(dpy);
	rwin = RootWindow(dpy, scr);

	xattr.background_pixel = xattr.border_pixel = BlackPixel(dpy, scr);
	xattr.colormap = DefaultColormap(dpy, scr);

	win = XCreateWindow(dpy, rwin, 0, 0, xsz, ysz, 0, CopyFromParent, InputOutput,
			DefaultVisual(dpy, scr), CWColormap | CWBackPixel | CWBorderPixel, &xattr);
	XSelectInput(dpy, win, ExposureMask | StructureNotifyMask | KeyPressMask);

	wm_del = XInternAtom(dpy, "WM_DELETE_WINDOW", True);
	XSetWMProtocols(dpy, win, &wm_del, 1);

	XStringListToTextProperty(&title, 1, &wname);
	XSetWMName(dpy, win, &wname);
	XFree(wname.value);

	chint.res_name = "xwrp";
	chint.res_class = "xwrp";
	XSetClassHint(dpy, win, &chint);

	if(!XShapeQueryExtension(dpy, &event_base, &error_base)) {
		fputs("Shape extension unsupported\n", stderr);
		XDestroyWindow(dpy, win);
		XCloseDisplay(dpy);
		return -1;
	}

	set_no_decoration(win);

	XpmCreatePixmapFromData(dpy, rwin, gnu_col, &pixmap, 0, 0);

	mask = XCreateBitmapFromData(dpy, rwin, (char*)gnu_mask_bits, gnu_mask_width, gnu_mask_height);
	XShapeCombineMask(dpy, win, ShapeBounding, 0, 0, mask, ShapeSet);

	XMapWindow(dpy, win);
	XFlush(dpy);
	
	return 0;
}

void clean_up(void) {
	XDestroyWindow(dpy, win);
	XCloseDisplay(dpy);
}

struct mwm_hints{
    uint32_t flags;
    uint32_t functions;
    uint32_t decorations;
    int32_t input_mode;
    uint32_t status;
};

#define MWM_HINTS_DEC	(1 << 1)
#define MWM_DECOR_NONE	0

void set_no_decoration(Window win) {
	Atom wm_hints;

	if((wm_hints = XInternAtom(dpy, "_MOTIF_WM_HINTS", True)) != None) {
		struct mwm_hints hints = {MWM_HINTS_DEC, 0, MWM_DECOR_NONE, 0, 0};
		XChangeProperty(dpy, win, wm_hints, wm_hints, 32, PropModeReplace, (unsigned char*)&hints, 4);
	}
}
